"""
region.py - Python module wrapper for code region
"""
from functools import partial as _partial, wraps as _wraps
from inspect import stack as _stack
from os.path import basename as _basename

from .string_handle import string_handle as _string_handle


class _Region:
    """
    An abstract base class that provides common functionality to wrap a code region.

    The subclasses itself and instances of the subclasses can be used as a context manager or as a decorator
    to automatically track the execution of code region and callable objects (e.g. function).
    """
    def __init__(self, func=None, deferred_wrap_callback=None) -> None:
        """
        Creates the instance of class that represents a traced code region.
        :param func: a callable object to wrap. If it is None, a wrapper creation will be deferred and can be done
                     using __call__() function for the object.
        :param deferred_wrap_callback: a callable object that will be called when wrapper is created if func is None.
        """
        self.__function = func
        self.__deferred_wrap_callback_function = deferred_wrap_callback

        if self.__function is None:
            self.__call_target = self.__wrap
        elif callable(self.__function):
            self.__call_target = self.__get_wrapper(self.__function)
            _wraps(self.__function, updated=())(self)
        else:
            raise TypeError('func must be a callable object or None.')

    def __get__(self, obj, objtype):
        return _wraps(self)(self.__get_wrapper(self.__function, obj))

    def __enter__(self) -> None:
        self.begin()

    def __exit__(self, *args) -> None:
        self.end()

    def __call__(self, *args, **kwargs):
        return self.__call_target(*args, **kwargs)

    def begin(self) -> None:
        """Marks the beginning of a code region."""
        raise NotImplementedError()

    def end(self) -> None:
        """Marks the end of a code region."""
        raise NotImplementedError()

    def __wrap(self, func):
        """
        Wraps a callable object.
        :param func: a callable object to wrap
        :return: a wrapper to trace the execution of the callable object
        """
        if callable(func):
            self.__function = func
        else:
            raise TypeError('Callable object is expected as a first argument.')

        self.__call_wrap_callback()

        return _wraps(self.__function)(self.__get_wrapper(self.__function))

    def __call_wrap_callback(self):
        """
        Call a callback for wrapper creation.
        """
        if callable(self.__deferred_wrap_callback_function):
            self.__deferred_wrap_callback_function(self.__function)

    def __get_wrapper(self, func, obj=None):
        """
        Returns a pure wrapper for a callable object.
        :param func: the callable object to wrap
        :param obj: an object to which the callable object is bound
        :return: the wrapper to trace the execution of the callable object
        """
        if not callable(func):
            raise TypeError('Callable object is expected to be passed.')

        def _function_wrapper(*args, **kwargs):
            """
            A wrapper to trace the execution of a callable object
            :param args: positional arguments of the callable object
            :param kwargs: keyword arguments of the callable object
            :return: result of a call of the callable object
            """
            self.begin()

            try:
                func_result = func(*args, **kwargs)
            finally:
                self.end()

            return func_result

        def _method_wrapper(*args, **kwargs):
            """
            A wrapper to trace the execution of a class method
            :param args: positional arguments of the class method
            :param kwargs: keyword arguments of the class method
            :return: result of a call of the class method
            """
            self.begin()

            try:
                func_result = func(obj, *args, **kwargs)
            finally:
                self.end()

            return func_result

        return _function_wrapper if obj is None or isinstance(func, staticmethod) else _method_wrapper


class _CallSite:
    """
    A class that represents a call site for a callable object.
    """
    CallerFrame = 1

    def __init__(self, frame_number: int) -> None:
        """
        Creates a call site.
        :param frame_number: relative frame number that should be used to extract the information about the call site
        """
        caller = _stack()[frame_number+1]
        self._filename = _basename(caller.filename)
        self._lineno = caller.lineno

    def filename(self):
        """Returns filename for the call site."""
        return self._filename

    def lineno(self):
        """Returns line number for the call site."""
        return self._lineno


class _NamedRegion(_Region):
    """
    An abstract base class that represents a named code region.
    """
    def __init__(self, func=None, name_creation_callback=None) -> None:
        """
        Creates the instance of class that represents a named code region.
        :param func: a name of the code region, a call site for the code region or a callable object (e.g. function) to
                     wrap.
                     If the call site object is passed, it is used as the initial choice for the region name. The name
                     that was derived based on call site object can be replaced with the name of the callable object if
                     it is passed in the future.
                     If the callable object is passed the name of this object is used as a name for the code region.
        """
        super().__init__(self.__get_function(func), _partial(_NamedRegion.__deferred_wrap_callback, self))

        self._name = self.__get_name(func)
        self.__name_creation_callback = name_creation_callback
        self.__is_final_name_determined = False
        self.__is_custom_name_specified = isinstance(func, str)

        final_name_can_be_determined_now = not (func is None or isinstance(func, _CallSite))
        if final_name_can_be_determined_now:
            self.__original_begin_func = None
            self.__determine_final_name()
        else:
            self.__original_begin_func = self.begin
            self.begin = self.__determine_final_name

    def __str__(self) -> str:
        return self._name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self._name}')"

    def begin(self) -> None:
        """Marks the beginning of a code region."""
        raise NotImplementedError()

    def end(self) -> None:
        """Marks the end of a code region."""
        raise NotImplementedError()

    def name(self):
        """Return the name of the code region."""
        return self._name

    def __determine_final_name(self):
        """Determines a final name of a code region"""
        self.__is_final_name_determined = True

        if callable(self.__name_creation_callback):
            self.__name_creation_callback(self._name)

        if self.__original_begin_func is not None:
            self.begin = self.__original_begin_func
            self.begin()

    def __deferred_wrap_callback(self, func):
        """Determines a final name of a code region if it has not been done before"""
        if not self.__is_final_name_determined:
            self._name = self.__get_name(func)
            self.__determine_final_name()
        elif not self.__is_custom_name_specified:
            raise RuntimeError('A custom name for a code region must be specified before'
                               ' _NamedRegion.__call__() can be called more than once.')

    @staticmethod
    def __get_function(func):
        """Returns the argument if it is callable, otherwise returns None"""
        return func if callable(func) else None

    @staticmethod
    def __get_name(func):
        """Returns appropriate code region name"""
        if func is None:
            return None

        if isinstance(func, str):
            return _string_handle(func)

        if isinstance(func, _CallSite):
            return _string_handle(f'{func.filename()}:{func.lineno()}')

        if hasattr(func, '__qualname__'):
            return _string_handle(func.__qualname__)

        if hasattr(func, '__name__'):
            return _string_handle(func.__name__)

        if hasattr(func, '__class__'):
            return _string_handle(f'{func.__class__.__name__}.__call__')

        raise ValueError('Cannot get the name for the code region.')



"""
event.py - Python module wrapper for ITT Event API
"""
from functools import partial as _partial

from pyitt.native import Event as _Event

from .region import _CallSite, _NamedRegion


class Event(_NamedRegion):
    """
    A class that represents Event.
    """
    def __init__(self, region=None):
        """
        Creates the instance of the class that represents ITT Event.
        :param region: a name of the event or a callable object (e.g. function) to wrap. If the callable object is
                       passed the name of this object is used as a name for the event.
        """
        self._event = None

        super().__init__(region, _partial(Event.__deferred_event_creation, self))

    def __deferred_event_creation(self, name) -> None:
        """Performs deferred creation of native Event."""
        self._event = _Event(name)

    def begin(self) -> None:
        """Marks the beginning of an event region."""
        self._event.begin()

    def end(self) -> None:
        """Marks the end of an event region."""
        self._event.end()


def event(region=None) -> Event:
    """
    Creates an Event instance.
    :param region: a name of the event or a callable object (e.g. function) to wrap. If the callable object is
                   passed the name of this object is used as a name for the event.
    :return: an Event instance
    """
    region = _CallSite(_CallSite.CallerFrame) if region is None else region
    return Event(region)



"""
string_handle.py - Python module wrapper for ITT String Handle API
"""
from pyitt.native import StringHandle as _StringHandle


def string_handle(string: str):
    """
    Creates a handle for a string.
    :param string: a string
    :return: the handle to the given string
    """
    return _StringHandle(string)



"""
domain.py - Python module wrapper for ITT Domain API
"""
from pyitt.native import Domain as _Domain


def domain(name=None):
    """
    Creates a domain with the given name.
    :param name: a name of the domain
    :return: a handle to the created domain with given name if the `name` is not None,
     otherwise, returns handle to default domain.
    """
    return _Domain(name)



"""
task.py - Python module wrapper for ITT Task API
"""
from pyitt.native import task_begin as _task_begin, task_end as _task_end
from pyitt.native import task_begin_overlapped as _task_begin_overlapped, task_end_overlapped as _task_end_overlapped

from .domain import domain as _domain
from .id import id as _id
from .region import _CallSite, _NamedRegion


class _Task(_NamedRegion):
    """
    An abstract base class that provides common functionality for subtypes that represent ITT Tasks.
    """
    def __init__(self, task=None, domain=None, id=None, parent=None) -> None:
        """
        Creates the instance of the class that represents ITT task.
        :param task: a name of the task or a callable object (e.g. function) to wrap. If the callable object is passed
                     the name of this object is used as a name for the task.
        :param domain: a task domain
        :param id: a task id
        :param parent: a parent task or an id of the parent
        """
        super().__init__(task)

        self._domain = self.__get_task_domain(domain)
        self._id = self.__get_task_id(id, self._domain)
        self._parent_id = self.__get_parent_id(parent)

    def __str__(self) -> str:
        return (f"{{ name: '{str(self._name)}', domain: '{str(self._domain)}',"
                f" id: {str(self._id)}, parent_id: {str(self._parent_id)} }}")

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}({repr(self._name)}, {repr(self._domain)},'
                f' {repr(self._id)}, {repr(self._parent_id)})')

    def domain(self):
        """Returns the domain of the task."""
        return self._domain

    def id(self):
        """Returns the id of the task."""
        return self._id

    def parent_id(self):
        """Returns the parent id for the task."""
        return self._parent_id

    def begin(self) -> None:
        """Marks the beginning of a task."""
        raise NotImplementedError()

    def end(self) -> None:
        """Marks the end of a task."""
        raise NotImplementedError()

    @staticmethod
    def __get_task_domain(original_domain):
        """Returns task domain"""
        if original_domain is None or isinstance(original_domain, str):
            return _domain(original_domain)

        return original_domain

    @staticmethod
    def __get_task_id(original_id, domain):
        """Returns task id for specified domain"""
        return _id(domain) if original_id is None else original_id

    @staticmethod
    def __get_parent_id(original_parent):
        """Returns parent id"""
        return original_parent.id() if isinstance(original_parent, task.__class__) else original_parent


class NestedTask(_Task):
    """
    A class that represents nested tasks.

    Nested tasks implicitly support a concept of embedded execution. This means that the call end() finalizes the
    most recent begin() call of the same or another nested task.
    """
    def begin(self) -> None:
        """Marks the beginning of a task."""
        _task_begin(self._domain, self._name, self._id, self._parent_id)

    def end(self) -> None:
        """Marks the end of a task."""
        _task_end(self._domain)


def nested_task(task=None, domain=None, id=None, parent=None):
    """
    Creates a nested task instance with the given arguments.
    :param task: a name of the task or a callable object
    :param domain: a task domain
    :param id: a task id
    :param parent: a parent task or an id of the parent
    :return: an instance of NestedTask
    """
    task = _CallSite(_CallSite.CallerFrame) if task is None else task
    return NestedTask(task, domain, id, parent)


class OverlappedTask(_Task):
    """
    A class that represents overlapped tasks.

    Execution regions of overlapped tasks may intersect.
    """
    def begin(self) -> None:
        """Marks the beginning of a task."""
        _task_begin_overlapped(self._domain, self._name, self._id, self._parent_id)

    def end(self) -> None:
        """Marks the end of a task."""
        _task_end_overlapped(self._domain, self._id)


def overlapped_task(task=None, domain=None, id=None, parent=None):
    """
    Creates an overlapped task instance with the given arguments.
    :param task: a name of the task or a callable object
    :param domain: a task domain
    :param id: a task id
    :param parent: a parent task or an id of the parent
    :return: an instance of OverlappedTask
    """
    task = _CallSite(_CallSite.CallerFrame) if task is None else task
    return OverlappedTask(task, domain, id, parent)


def task(task=None, domain=None, id=None, parent=None, overlapped=False):
    """
    Creates a task instance with the given arguments.
    :param task: a name of the task or a callable object
    :param domain: a task domain
    :param id: a task id
    :param parent: a parent task or an id of the parent
    :param overlapped: determines if the created task should be an instance of OverlappedTask class
                       or NestedTask class
    :return: a task instance
    """
    task = _CallSite(_CallSite.CallerFrame) if task is None else task
    return OverlappedTask(task, domain, id, parent) if overlapped else NestedTask(task, domain, id, parent)



"""
Python binding to Intel Instrumentation and Tracing Technology (ITT) API.

This module provides a convenient way to mark up the Python code for further performance analysis using performance
analyzers from Intel like Intel VTune or others.
"""

from pyitt.native import Domain, Id, StringHandle
from pyitt.native import task_begin, task_end, task_begin_overlapped, task_end_overlapped
from .collection_control import detach, pause, resume, active_region, paused_region, ActiveRegion, PausedRegion
from .event import event, Event
from .domain import domain
from .id import id
from .string_handle import string_handle
from .task import NestedTask, OverlappedTask, task, nested_task, overlapped_task
from .thread_naming import thread_set_name



"""
id.py - Python module wrapper for ITT ID API
"""
from pyitt.native import Id as _Id


def id(domain):
    """
    Creates a unique identifier.
    :param domain: a domain that controls the creation of the identifier
    :return: an instance of the identifier
    """
    return _Id(domain)



"""
collection_control.py - Python module wrapper for ITT Collection Control API
"""
from pyitt.native import detach as _detach, pause as _pause, resume as _resume

from .region import _Region


class _CollectionRegion(_Region):
    """
    An abstract base class that provides common functionality for subclasses that represent paused/resumed collection
    regions.
    """
    def __init__(self, func=None, activator=None):
        """
        Creates a collection region.
        :param func: a callable object that represents the collection region, e.g. function.
        :param activator: a callable object that determines if the region is active or not. The callable object should
                          not take any arguments and should return True if the region is active, otherwise should return
                          False. If the region is active, _CollectionRegion performs _begin()/_end() calls to inform
                          subclasses about the start and end of the region. Otherwise, it does nothing.
        """
        super().__init__(func)
        self.activator = activator
        self.__is_paired_call_needed = False

    def _begin(self):
        raise NotImplementedError()

    def _end(self):
        raise NotImplementedError()

    def begin(self):
        """Marks the beginning of a collection region."""
        if callable(self.activator):
            activator_state = self.activator()
            self.__is_paired_call_needed = activator_state

            if activator_state:
                self._begin()
        else:
            self.__is_paired_call_needed = True
            self._begin()

    def end(self):
        """Marks the end of a collection region."""
        if self.__is_paired_call_needed:
            self._end()


def detach() -> None:
    """Detach collection of profiling data."""
    _detach()


def pause() -> None:
    """Pause collection of profiling data."""
    _pause()


def resume() -> None:
    """Resume collection of profiling data."""
    _resume()


class ManualCollectionRegionActivator:
    """
    A class that provides ability to activate/deactivate paused/resumed regions.
    """
    INACTIVE = 0
    ACTIVE = 1

    def __init__(self, state=ACTIVE):
        """
        Creates an activator.
        :param state: sets the initial state of activator.
        """
        self._state = state

    def __call__(self):
        return self._state == self.ACTIVE

    def activate(self):
        """Activates the region."""
        self._state = self.ACTIVE

    def deactivate(self):
        """Deactivates the region."""
        self._state = self.INACTIVE


class ActiveRegion(_CollectionRegion):
    """
    A class that represents resumed collection region.

    It allows to collect profiling only for this region. The collection of profiling data have to be run in
    Start Paused mode.
    """
    def __init__(self, func=None, activator=ManualCollectionRegionActivator()):
        """
        Creates an instance of the class that represents resumed collection region.
        :param func: a callable object that represents the collection region, e.g. function.
        :param activator: a callable object that determines if the region is active or not. The callable object should
                          not take any arguments and should return True if the region is active, otherwise should return
                          False. If the region is active, a call of begin() method of the instance will resume
                          the collection of profiling data and a call of end() method will pause the collection again.
                          Otherwise, these calls do nothing.
        """
        super().__init__(func, activator)

    def _begin(self):
        resume()

    def _end(self):
        pause()


def active_region(func=None, activator=ManualCollectionRegionActivator()):
    """
    Creates a resumed collection region with the given arguments.
    :param func: a callable object that represents the collection region, e.g. function.
    :param activator: a callable object that determines if the region is active or not. The callable object should
                      not take any arguments and should return True if the region is active, otherwise should return
                      False. If the region is active, a call of begin() method of the instance will resume
                      the collection of profiling data and a call of end() method will pause the collection again.
                      Otherwise, these calls do nothing.
    :return: an instance of ActiveRegion.
    """
    return ActiveRegion(func, activator)


class PausedRegion(_CollectionRegion):
    """
    A class that represents paused collection region.

    An instance of this class allows to disable the collection of profiling data for the code region that is not
    interested.
    """
    def __init__(self, func=None, activator=ManualCollectionRegionActivator()):
        """
        Creates an instance of the class that represents paused collection region.
        :param func: a callable object that represents the collection region, e.g. function.
        :param activator: a callable object that determines if the region is active or not. The callable object should
                          not take any arguments and should return True if the region is active, otherwise should return
                          False. If the region is active, a call of begin() method for the instance will pause
                          the collection of profiling data and a call of end() method will resume the collection again.
                          Otherwise, these calls do nothing.
        """
        super().__init__(func, activator)

    def _begin(self):
        pause()

    def _end(self):
        resume()


def paused_region(func=None, activator=ManualCollectionRegionActivator()):
    """
    Creates a paused collection region with the given arguments.
    :param func: a callable object that represents the collection region, e.g. function.
    :param activator: a callable object that determines if the region is active or not. The callable object should
                      not take any arguments and should return True if the region is active, otherwise should return
                      False. If the region is active, a call of begin() method for the instance will pause
                      the collection of profiling data and a call of end() method will resume the collection again.
                      Otherwise, these calls do nothing.
    :return: an instance of PausedRegion.
    """
    return PausedRegion(func, activator)



"""
thread_naming.py - Python module wrapper for ITT Thread Naming API
"""
from pyitt.native import thread_set_name as _thread_set_name


def thread_set_name(name: str):
    """
    Sets a thread name of calling thread.
    :param name: the thread name
    """
    _thread_set_name(name)



"""
setup.py - Python module to install pyitt package
"""
import os
import sys

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from subprocess import run  # pylint: disable=C0411


def get_environment_flag(name):
    """
    Get an environment variable and check its value.
    :param name: the environment variable name
    :return: None if the environment variable is not set. False if the environment variable is set and has one of the
             following values: `''`, `0`, `n`, `no` or `false`. Otherwise, return True.
    """
    flag_value = os.environ.get(name, None)
    return (flag_value and flag_value.lower() not in ('0', 'n', 'no', 'false')) if flag_value is not None else None


IS_64_ARCHITECTURE = sys.maxsize > 2 ** 32

# Check if custom location for ITT API source code is specified
ITT_DEFAULT_DIR = 'ittapi'
itt_dir = os.environ.get('PYITT_ITT_API_SOURCE_DIR', None)
itt_dir = itt_dir if itt_dir else ITT_DEFAULT_DIR

assert os.path.exists(itt_dir), 'The specified directory with ITT API source code does not exist.'
assert itt_dir != ITT_DEFAULT_DIR or len(os.listdir(itt_dir)), \
    (f'The specified directory with ITT API source code ({itt_dir}) is empty.\n'
     f'Make sure that submodules are checked out as well using following command:\n'
     f'git submodule update --init --recursive')

# Check if IPT support is requested
build_itt_with_ipt_support = get_environment_flag('PYITT_BUILD_WITH_ITT_API_IPT_SUPPORT')
build_itt_with_ipt_support = build_itt_with_ipt_support if build_itt_with_ipt_support is not None else True

itt_source = [os.path.join(itt_dir, 'src', 'ittnotify', 'ittnotify_static.c')]
itt_include_dirs = [os.path.join(itt_dir, 'include')]
itt_license_files = [os.path.join(itt_dir, 'LICENSES', 'BSD-3-Clause.txt')] if itt_dir == ITT_DEFAULT_DIR else []

if build_itt_with_ipt_support:
    itt_compiler_flags = ['-DITT_API_IPT_SUPPORT']
    if sys.platform == 'win32':
        ITT_PTMARK_SOURCE = 'ittptmark64.asm' if IS_64_ARCHITECTURE else 'ittptmark32.asm'
    else:
        ITT_PTMARK_SOURCE = 'ittptmark64.S' if IS_64_ARCHITECTURE else 'ittptmark32.S'
    itt_extra_objects = [os.path.join(itt_dir, 'src', 'ittnotify', ITT_PTMARK_SOURCE)]
else:
    itt_compiler_flags = []
    itt_extra_objects = []

pyitt_license_files = ['LICENSE']
pyitt_native_sources = ['pyitt.native/extensions/python.cpp',
                        'pyitt.native/extensions/string.cpp',
                        'pyitt.native/collection_control.cpp',
                        'pyitt.native/domain.cpp',
                        'pyitt.native/event.cpp',
                        'pyitt.native/id.cpp',
                        'pyitt.native/string_handle.cpp',
                        'pyitt.native/task.cpp',
                        'pyitt.native/thread_naming.cpp',
                        'pyitt.native/pyitt.cpp']

pyitt_native_compiler_args = ['/std:c++20' if sys.platform == 'win32' else '-std=c++20']
if build_itt_with_ipt_support:
    pyitt_native_compiler_args.append('-DPYITT_BUILD_WITH_ITT_API_IPT_SUPPORT=1')

pyitt_native = Extension('pyitt.native',
                         sources=itt_source + pyitt_native_sources,
                         include_dirs=itt_include_dirs,
                         extra_compile_args=itt_compiler_flags + pyitt_native_compiler_args,
                         extra_objects=itt_extra_objects)


class NativeBuildExtension(build_ext):  # pylint: disable=R0903
    """
    A class that implements the build extension to compile pyitt.native module.
    """
    def build_extension(self, ext) -> None:
        """
        Build native extension module
        :param ext: the extension to build
        """
        if ext.name == 'pyitt.native' and self.compiler.compiler_type == 'msvc':
            # Setup asm tool
            as_tool = 'ml64.exe' if IS_64_ARCHITECTURE else 'ml.exe'
            as_ext = '.asm'

            if hasattr(self.compiler, 'initialized') and hasattr(self.compiler, 'initialize'):
                if not self.compiler.initialized:
                    self.compiler.initialize()

            as_path = os.path.dirname(self.compiler.cc) if hasattr(self.compiler, 'cc') else ''

            # Extract asm files from extra objects
            # pylint: disable=W0106
            asm_files = [filename for filename in ext.extra_objects if filename.lower().endswith(as_ext)]
            [ext.extra_objects.remove(filename) for filename in asm_files]

            # Create temp directories
            [os.makedirs(os.path.join(self.build_temp, os.path.dirname(filename)), exist_ok=True)
             for filename in asm_files]

            # Generate target names
            src_dir = os.path.dirname(__file__)
            obj_asm_pairs = [(os.path.join(self.build_temp, os.path.splitext(filename)[0]) + '.obj',
                              os.path.join(src_dir, filename)) for filename in asm_files]
            # Compile
            [run([os.path.join(as_path, as_tool), '/Fo', obj_file, '/c', asm_file], check=True)
             for obj_file, asm_file in obj_asm_pairs]

            [ext.extra_objects.append(obj_file) for obj_file, _ in obj_asm_pairs]

        build_ext.build_extension(self, ext)


setup(name='pyitt',
      version='1.1.0',
      description='ITT API bindings for Python',
      packages=['pyitt'],
      ext_modules=[pyitt_native],
      license_files=pyitt_license_files + itt_license_files,
      cmdclass={'build_ext': NativeBuildExtension} if build_itt_with_ipt_support else {},
      zip_safe=False)
